<?php

declare(strict_types=1);

namespace mirzaev\yii2\sessions\arangodb;

// Фреймворк Yii
use yii;
use yii\di\Instance;
use yii\base\InvalidConfigException;

// Модуль ArangoDB для фреймворка Yii
use mirzaev\yii2\arangodb\Connection,
    mirzaev\yii2\arangodb\Query;

// Встроенные библиотеки
use Exception;

/**
 * Реализация DbSession под ArangoDB
 *
 * @property Connection|array|string $database Идентификатор компонента с базой данных (Component ID)
 * @property string $document Название документа для хранения данных сеанса (Document name for storing session data)
 * @property array $buffer Буфер данных для записи в документ сессии (Data buffer for write to session document)
 *
 * @method public init() Инициализация компонента (Component initialization)
 * @method public openSession() Открыть сессию (Open session)
 * @method public close() Закрыть сессию (Close session)
 * @method public regenerateID() Обновление сессии (Update session)
 * @method public readSession() Прочитать сессию (Read session)
 * @method public searchSession() Найти сессию (Search session)
 * @method public initSession() Инициализировать сессию (Initialize session)
 * @method public writeSession() Записать сессию (Write session)
 * @method public destroySession() Удалить сессию (Delete session)
 * @method public gcSession() Удалить неиспользуемые данные (Delete garbage)
 * @method protected getReadQuery() Генерация запроса для чтения сеанса (Generating a query to read a session)
 * @method protected genBuffer() Генерация массива данных для записи (Generating an array of data to writing)
 * @method protected typecastFields() Конвертация для отправки в PDO (Convertation  for sending to PDO)
 *
 * @example ./Migrations/create_sessions_collection.php
 *
 * @uses $this->gcSession() Do not call this method directly
 * @uses $this->openSession() Do not call this method directly
 * @uses $this->readSession() Do not call this method directly
 * @uses $this->writeSession() Do not call this method directly
 * @uses $this->destroySession() Do not call this method directly
 *
 * @see yii\web\DbSession Наследует, копирует (Inherit, copy)
 *
 * @package yii2\ArangoDB
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 *
 * @since 2.0
 */
final class ArangoDbSession extends MultiFieldSession
{

    /**
     * Идентификатор компонента с базой данных (Component ID)
     */
    public Connection|array|string $database = 'arangodb';

    /**
     * Название документа для хранения данных сеанса (Document name for storing session data)
     */
    public string $document = 'sessions';

    /**
     * Буфер данных для записи в документ сессии (Data buffer for write to session document)
     */
    protected array $buffer = [];


    /**
     * Инициализация компонента (Component initialization)
     *
     * @throws InvalidConfigException if $this->database is invalid.
     */
    public function init(): Connection
    {
        return $this->database = Instance::ensure($this->database, Connection::class);
    }

    /**
     * Открыть сессию (Open session)
     *
     * @param string $path Путь для сохранения сессии (Path to save session)
     * @param string $name Имя сессии (Session name)
     *
     * @return bool Статус открытия сессии (Session opening status)
     */
    public function openSession($path, $name): bool
    // public function openSession(string $path, string $name): bool
    {
        if ($this->getUseStrictMode()) {

            $id = $this->getId();

            if (!$this->getReadQuery($id)->exists()) {
                // This session id does not exist, mark it for forced regeneration
                $this->_forceRegenerateId = $id;
            }
        }

        return parent::openSession($path, $name);
    }

    /**
     * Закрыть сессию (Close session)
     *
     * Закрывает текущую сессию и сохраняет данные
     * Ends the current session and store session data
     */
    public function close(): void
    {
        if ($this->getIsActive()) {
            // prepare writeCallback fields before session closes

            $this->buffer = $this->composeFields();

            YII_DEBUG ? session_write_close() : @session_write_close();
        }
    }

    /**
     * Обновление сессии (Update session)
     *
     * Метод не создаст эффект пока сессия неактивна (getIsActive()). Убедитесь, что открыли (open()) перед вызовом
     * This method has no effect when session is not [[getIsActive()|active]]. Make sure to call [[open()]] before calling it.
     *
     * @param bool $rewrite Перезаписать прошлую сессию (Rewrite old session)
     *
     * @see $this->open()
     * @see $this->getIsActive()
     *
     * @tutorial https://secure.php.net/session_regenerate_id
     */
    public function regenerateID($rewrite = false): void
    // public function regenerateID(bool $rewrite = false): void
    {
        // Инициализация
        $oldID = session_id();

        if (empty($oldID)) {
            // if no session is started, there is nothing to regenerate

            return;
        }

        parent::regenerateID(false);

        $newID = session_id();

        // if session id regeneration failed, no need to create/update it.
        if (empty($newID)) {
            yii::warning('Failed to generate new session ID', __METHOD__);

            return;
        }

        $row = (new Query())
            ->in($this->document)
            ->where(['hash' => $oldID])
            ->one($this->database);

        if ($row) {
            if ($rewrite) {
                // Обновить

                (new Query())
                    ->in($this->document)
                    ->where(['hash' => $oldID])
                    ->update($this->genBuffer($newID), db: $this->database);
            } else {
                // Создать

                // Инициализация
                $new = $row;
                $new['id'] = $newID;

                (new Query())
                    ->in($this->document)
                    ->insert($new, db: $this->database);
            }
        } else {
            // shouldn't reach here normally
            (new Query())
                ->in($this->document)
                ->insert($this->genBuffer($newID), db: $this->database);

            // (new Query())
            //     ->in($this->document)
            //     ->insert($this->composeFields($newID, ''), db: $this->database); !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        }
    }

    /**
     * Прочитать сессию (Read session)
     *
     * @param string $id Идентификатор сессии (Session ID)
     *
     * @return string Данные сессии (The session data)
     */
    public function readSession($hash): string
    // public function readSession(string $hash): string
    {
        $query = $this->getReadQuery($hash);

        if ($this->readCallback !== null) {
            $data = $query->one($this->database);

            return $data ? $this->extractData($data) : '';
        }

        $data = $query->select(['data' => 'data'])->one($this->database);

        return $data['data'] ?? '';
    }

    /**
     * Найти сессию (Search session)
     *
     * @param string $hash Идентификатор сессии (Session ID)
     *
     * @return array|null Сессия, если найдена (Session, if founded)
     */
    public function searchSession(string $hash): ?array
    {
        if (empty($hash)) {
            return false;
        }

        $session = (new Query())
            ->in($this->document)
            ->where([
                'hash' => $hash,
                'status' => 'active'
            ])
            ->one($this->database);

        return $session ? $session : null;
    }

    /**
     * Инициализировать сессию (Initialize session)
     *
     * @param string $hash Идентификатор сессии (Session ID)
     * @param string|null $data Данные сессии (Session data)
     *
     * @return array Сессия или статус обновления (Session or update status)
     */
    public function initSession(string $hash, string|null $data = null): array|bool|null
    {
        if (empty($hash)) {
            return null;
        }

        if ($session = $this->searchSession($hash)) {
            if (isset($data, $session['data']) && $session['data'] !== $data) {
                // Необходима синхронизация (данные не идентичны)

                return (new Query())
                    ->in($this->document)
                    ->where(['hash' => $session['hash']])
                    ->update(['data' => $data], db: $this->database);
            }

            return $session;
        }

        $query = (new Query())
            ->in($this->document)
            ->insert($this->genBuffer($hash, $data), db: $this->database);

        return $query ? $query : null;
    }

    /**
     * Записать сессию (Write session)
     *
     * @param string $id Идентификатор сессии (Session ID)
     * @param string $data Данные сессии (Session data)
     *
     * @return bool Статус записи сессии (Session writing status)
     */
    public function writeSession($id, $data): bool
    // public function writeSession(string $id, string $data): bool
    {
        if ($this->getUseStrictMode() && $id === $this->_forceRegenerateId) {
            //Ignore write when forceRegenerate is active for this id

            return true;
        }

        // exception must be caught in session write handler
        // https://secure.php.net/manual/en/function.session-set-save-handler.php#refsect1-function.session-set-save-handler-notes
        try {
            // ensure backwards compatability (fixed #9438)
            if ($this->writeCallback && !$this->buffer) {
                $this->buffer = $this->composeFields();
            }

            // ensure data consistency
            if (isset($this->buffer['data'])) {

                $_SESSION = $this->buffer['data'];
            } else {

                $this->buffer['data'] = $data;
            }

            $this->genBuffer($id, $data);

            // Конвертация или сериализация, не пойму
            $this->buffer = $this->typecastFields($this->buffer);

            // Инициализация сессии
            $this->initSession($id, $data);
        } catch (Exception $e) {
            yii::$app->errorHandler->handleException($e);

            return false;
        }

        return true;
    }

    /**
     * Закрыть сессию (Close session)
     *
     * @return bool Статус закрытия сессии (Session closing status)
     *
     * @todo Разобраться почему вызывается destroySession и избавиться от перезаписи (сделать обновление существующей)
     */
    public function closeSession(): bool
    {
        // try {
        //     $this->buffer['status'] = 'inactive';

        //     (new Query())
        //         ->in($this->document)
        //         ->insert($this->buffer, db: $this->database);
        // } catch (Exception $e) {
        //     yii::$app->errorHandler->handleException($e);

        //     return false;
        // }

        return true;
    }



    /**
     * Удалить сессию (Delete session)
     *
     * @param string $hash Идентификатор сессии (Session ID)
     *
     * @return bool Статус удаления сессии (Session deleting status)
     */
    public function destroySession($hash): bool
    // public function destroySession(string $hash): bool
    {
        return (new Query())
            ->in($this->document)
            ->where(['hash' => $hash])
            ->remove(db: $this->database);
    }

    /**
     * Удалить неиспользуемые данные (Delete garbage)
     *
     * Очищает от сессий чей срок хранения истёк
     * Clears sessions whose retention period has expired
     *.
     * @param int $max Периодичность очистки в секундах (Cleaning frequency in seconds)
     *
     * @return bool Статус удаления устаревших сессий (Deletion status of obsolete sessions)
     */
    public function gcSession($max): bool
    // public function gcSession(int $max): bool
    {
        return (new Query())
            ->in($this->document)
            ->where([
                [
                    [
                        'to' => time()
                    ],
                    'operator' => '<='
                ]
            ])
            ->remove(db: $this->database);
    }

    /**
     * Генерация запроса для чтения сессии (Generating a query to read a session)
     *
     * @param string $hash Идентификатор сессии (Session ID)
     *
     * @return Query Инстанция запроса на чтение (Read query instance)
     */
    protected function getReadQuery(string $hash): Query
    {
        return (new Query())
            ->in($this->document)
            ->where([
                [
                    [
                        'to' => time()
                    ],
                    'operator' => '>='
                ],
                [
                    [
                        'hash' => $hash,
                    ],
                    [
                        'status' => 'active'
                    ]
                ]
            ]);
    }

    /**
     * Генерация массива данных для записи (Generating an array of data to writing)
     *
     * @param string $hash Идентификатор сессии (Session ID)
     * @param string $data Данные сессии (Session data)
     *
     * @return array Обработанный буфер (Handled buffer)
     */
    protected function genBuffer(string $hash, string|null $data = null): array
    {
        return $this->buffer = array_merge(
            $this->buffer,
            $this->composeFields($hash, $data),
            [
                'from' => time(),
                'to' => time() + $this->getTimeout(),
                'status' => 'active'
            ]
        );
    }

    /**
     * Конвертация для отправки в PDO (Convertation  for sending to PDO)
     *
     * Default implementation casts field `data` to `\PDO::PARAM_LOB`.
     * You can override this method in case you need special type casting.
     *
     * @param array $data Данные для конвертации (Data for convertation) [name => value]
     *
     * @return array Конвертированные данные (Converted data)
     */
    protected function typecastFields(array $data): array
    {
        return $data;
    }
}
